import numpy as np
from tabulate import tabulate


class Instance:

    def __init__(self, values, attributes: dict = None):
        self.dataset = None
        self.data = np.array(values)
        # Estructura que almacene la informacion de los atributos(nombre, si es label, etc)
        # TODO: Ver si crear clase
        self.attributes = attributes

    def get_number_attributes(self):
        return len(self.attributes)

    def get_number_classes(self):
        cont = 0
        for x in self.attributes:
            if self.attributes[x] == 1:
                cont += 1
        return cont

    def get_attribute_by_index(self, index: int):
        return self.data.item(index)

    def get_attribute_by_name(self, attribute: str):
        index = list(self.attributes.keys()).index(attribute)
        return self.data.item(index)

    def set_attribute_by_index(self, index: int, value):
        self.data[index] = value

    def set_attribute_by_name(self, attribute: str, value):
        index = list(self.attributes.keys()).index(attribute)
        self.data[index] = value

    def set_dataset(self, dataset):
        self.dataset = dataset

    def add_attribute(self, name, position=0, value=0):
        if position is None:
            position = len(self.data)
        # TODO: Habria que ordenar el diccionario para poder poner el atributo en la posicion que se quiera
        self.attributes.insert(position, name)
        self.data = np.insert(self.data, position, value)

    def delete_attribute(self, position):
        pass

    def delete_attribute(self, name):
        pass

    def show_instance(self):
        # TODO: Check

        table = []

        table.append(list(self.data))

        # table = [['col 1', 'col 2', 'col 3', 'col 4'], [1, 2222, 30, 500], [4, 55, 6777, 1]]
        # print(tabulate(table, headers='firstrow', tablefmt='fancy_grid'))
        # print(tabulate([key], tablefmt="grid"))
        print(tabulate(table, tablefmt="grid", numalign="center"))



from sklearn.metrics import classification_report, hamming_loss

from classifier.knn_classifier import KNNClassifier
from datasets.load_dataset import load_dataset
from transformation.mimlTOmi.binary_relevance import BinaryRelevanceTransformation

dataset_train = load_dataset("../datasets/miml_birds_random_80train.arff", delimiter="'")
dataset_test = load_dataset("../datasets/miml_birds_random_20test.arff", delimiter="'")

binary_relevance_train = BinaryRelevanceTransformation(dataset_train)
binary_relevance_test = BinaryRelevanceTransformation(dataset_test)

datasets_train = binary_relevance_train.transform_dataset()
datasets_test = binary_relevance_test.transform_dataset()

classifiers = []

for i in range(len(datasets_train)):
    knn_classifier = KNNClassifier(k=5)
    print("Train attrib: ", datasets_train[i][0])
    print("Train labels: ", datasets_train[i][1])
    knn_classifier.train(datasets_train[i][0], datasets_train[i][1])
    classifiers.append(knn_classifier)

y_preds = []
y_tests = []
for i in range(dataset_train.get_number_labels()):
    y_pred = []
    for j in range(len(datasets_train[i][0])):
        y_pred.append(classifiers[i].predict(datasets_test[i][0][j]))
    y_preds.append(y_pred)
    y_tests.append(datasets_test[i][1])

# Evaluación del modelo
print("Reporte de clasificación:\n", classification_report(y_tests, y_preds, zero_division=0))

# print("Y TEST:",y_test)
# print("Y Pred:",y_pred)

print('Hamming Loss: ', round(hamming_loss(y_tests, y_preds), 2))



from datasets.load_dataset import load_dataset

dataset = load_dataset("../datasets/toy.csv")

dataset.show_bag("bag2")

dataset.add_instance("bag2", [12, 22, 13])

dataset.show_bag("bag2")

# TODO: Añadir ejemplo de set_attribute, add_attribute y add_bag

dataset.describe()



import numpy as np

from data.miml_dataset import MIMLDataset


class MinMaxTransformation:
    """
    Class that performs a minmax transformation to convert a MIMLDataset class to numpy ndarrays.
    """

    def __init__(self, dataset: MIMLDataset):
        self.dataset = dataset

    def transform_dataset(self):
        """
        Transform the dataset to multilabel dataset converting each bag into a single instance with the min and max
        value of each attribute as two new attributes.

        Returns
        -------

        X : {numpy ndarray} of shape (number of instances, number of attributes)
        Training vector

        Y : {numpy ndarray} of shape (number of instances, number of labels)
        Target vector relative to X.

        """

        x = np.empty(shape=(len(self.dataset.data.keys()), self.dataset.get_number_attributes() * 2))
        y = np.empty(shape=(len(self.dataset.data.keys()), self.dataset.get_number_labels()))
        count = 0
        for keys, pattern in self.dataset.data.items():
            min_values = np.min(pattern[0], axis=0)
            max_values = np.max(pattern[0], axis=0)
            x[count] = min_values
            x[count + 1] = max_values
            y[count] = pattern[1]
            count += 2

        return x, y

    def transform_instance(self, key):
        """
        Transform the instances of a bag to a multilabel instance

        Parameters
        ----------
        key : string
            Key of the bag to be transformed to multilabel instance

        Returns
        -------
        instance : tuple
        Tuple of numpy ndarray with attribute values and labels

        """

    # TODO: Implementarlo



import os

import numpy as np

from data.bag import Bag
from data.instance import Instance
from data.miml_dataset import MIMLDataset


def load_dataset(file, delimiter="\""):
    """
    Function to load a dataset

    Parameters
    ----------
    file : string
        Path of the dataset file
    """
    if file[-4:] == ".csv":
        return load_dataset_csv(file)
    elif file[-5:] == ".arff":
        return load_dataset_arff(file, delimiter)
    else:
        print("Error")
        # TODO: Control de errores


def load_dataset_csv(file, header=0):
    """
    Function to load a dataset in csv format

    Parameters
    ----------
    file : string
        Path of the dataset file
    """
    # dataset = pd.read_csv(file, header=0)
    # TODO: Hay que ver como diferenciar los atributos de las labels
    # TODO: Si no se puede implementar la funcionalidad de pandas "[]"
    # TODO: y poner atributos y labels como parametros opcionales quizas

    dataset = MIMLDataset()
    csv_file = open(file)

    file_name = os.path.basename(file)
    dataset.set_name(os.path.splitext(file_name)[0])

    # TODO: Hacer que se pueda pasar por parametro
    num_labels = int(csv_file.readline().replace("\n", ""))

    header_line = csv_file.readline().replace("\n", "").split(",")
    attributes_name = header_line[1:-num_labels]
    # dataset.set_attributes(attributes_name)
    labels_name = header_line[-num_labels:]
    # dataset.set_labels(labels_name)
    attributes = dict()
    for x in attributes_name:
        attributes[x] = 0
    for y in labels_name:
        attributes[y] = 1

    dataset.set_attributes(attributes)
    for line in csv_file:

        data = line.split(",")

        key = data[0]

        values = np.array([float(i) for i in data[1:-num_labels]], ndmin=2)
        labels = np.array([int(i) for i in data[-num_labels:]])

        instance = Instance(values + labels, attributes)

        if key not in dataset.data:
            bag = Bag(instance, key)
            dataset.add_bag(bag)
        else:
            dataset.get_bag(key).add_instance(instance)

    return dataset


def load_dataset_arff(file, delimiter="\""):
    """
    Function to load a dataset in arff format

    Parameters
    ----------
    file : string
        Path of the dataset file
    """
    dataset = MIMLDataset()
    arff_file = open(file)
    attributes_name = []
    labels_name = []
    flag = 0
    for line in arff_file:

        # Comprobamos que la cadena no contenga espacios en blanco a la izquierda ni que sea vacía
        line = line.lstrip()

        if not line or line.startswith("%") or line.startswith("@data"):
            continue

        if line.startswith("@"):

            if line.startswith("@relation"):
                dataset.set_name(line[line.find(" ") + 1:])
            elif line.startswith("@attribute bag relational"):
                flag = 1
            elif line.startswith("@end bag"):
                flag = 2
            elif flag == 1:
                attributes_name.append(line.split(" ")[1])
            elif flag == 2:
                labels_name.append(line.split(" ")[1])

        else:
            # Eliminanos el salto de línea del final de la cadena
            line = line.strip("\n")

            # Asumimos que el primer elemento de cada instancia es el identificador de la bolsa
            key = line[0:line.find(",")]

            # Empiezan los datos de la bolsa cuando encontremos la primera '"' y terminan con la segunda '"'
            line = line[line.find(delimiter) + 1:]
            values = line[:line.find(delimiter, 2)]
            # Separamos los valores por instancias de la bolsa
            values = values.split("\\n")

            # El resto de la cadena se trata de las etiquetas
            labels = line[line.find(delimiter, 2) + 2:]
            labels_values = np.array([int(i) for i in labels.split(",")], ndmin=2)
            # print("Labels: ", labels)

            values_list = []
            for v in values:
                values_instance = np.array([float(i) for i in v.split(',')], ndmin=2)
                instance = Instance(np.hstack((values_instance, labels_values)))

                if key not in dataset.data:
                    bag = Bag(instance, key)
                    dataset.add_bag(bag)
                else:
                    dataset.get_bag(key).add_instance(instance)

    attributes = dict()
    for x in attributes_name:
        attributes[x] = 0
    for y in labels_name:
        attributes[y] = 1

    dataset.set_attributes(attributes)
    return dataset



import numpy as np

from data.miml_dataset import MIMLDataset


class BinaryRelevanceTransformation:
    """
    Class that performs a binary relevance transformation to convert a MIMLDataset class to numpy ndarrays.
    """

    def __init__(self, dataset: MIMLDataset):
        self.dataset = dataset

    def transform_dataset(self):
        """
        Transform the dataset to multilabel dataset converting each bag into a single instance being the value of each
        attribute the mean value of the instances in the bag.

        Returns
        -------

        X : {numpy ndarray} of shape (number of instances, number of attributes)
        Training vector

        Y : {numpy ndarray} of shape (number of instances, number of labels)
        Target vector relative to X.

        """
        datasets = []
        # TODO: Optimizar
        # x = np.zeros(shape=(self.dataset.get_number_bags(), self.dataset.get_number_attributes()))
        xs = []
        y = np.empty(shape=(self.dataset.get_number_bags(), 1))
        ys = []
        for i in range(self.dataset.get_number_labels()):
            ys.append(y.copy())
        count = 0
        for keys, pattern in self.dataset.data.items():
            # print("-------------------")
            xs.append(pattern[0])
            for i in range(self.dataset.get_number_labels()):
                # print(pattern[1][i])
                # print(ys)
                ys[i][count] = pattern[1][i]
            count += 1
        # xs = np.array(xs)

        for i in ys:
            datasets.append([xs, i])

        return datasets

    def transform_instance(self, key):
        """
        Transform the instances of a bag to a multilabel instance

        Parameters
        ----------
        key : string
            Key of the bag to be transformed to multilabel instance

        Returns
        -------
        instance : tuple
        Tuple of numpy ndarray with attribute values and labels

        """
        return 0

    # TODO: Implementarlo



from sklearn.metrics import accuracy_score
from sklearn.neighbors import KNeighborsClassifier

import classifier.abstract_classifier


class KNNClassifier(classifier.abstract_classifier.AbstractClassifier):

    def __init__(self, k=3):
        self.k = k
        self.model = None

    def train(self, training_data, training_labels):
        self.model = KNeighborsClassifier(n_neighbors=self.k)
        self.model.fit(training_data, training_labels)

    def predict(self, test_data):
        return self.model.predict(test_data)

    def evaluate(self, test_data, test_labels):
        predictions = self.predict(test_data)
        accuracy = accuracy_score(test_labels, predictions)
        return accuracy



import numpy as np
from tabulate import tabulate

from data.bag import Bag


class MIMLDataset:
    """"
    Class to manage MIML data obtained from datasets
    """

    def __init__(self) -> None:

        # TODO: Si dataset leido en csv, el nombre poner el del archivo
        self.name = "undefined"
        self.attributes = dict()
        self.data = dict()

    def set_name(self, name):
        """
        Set function for dataset name

        Parameters
        ----------
        name : string
            Name of the dataset
        """
        self.name = name

    def get_name(self):
        """
         Get function for dataset name

        Returns
        ----------
        name : string
            Name of the dataset
        """
        return self.name

    def set_attributes(self, attributes):
        """
        Set function for dataset attributes name

        Parameters
        ----------
        attributes : List of string
            List of the attributes name of the dataset
        """
        self.attributes = attributes

    def get_attributes(self):
        """
        Get function for dataset attributes name

        Returns
        ----------
        attributes : List of strings
            Attributes name of the dataset
        """
        return self.attributes

    def get_number_attributes(self):
        """
        Get numbers of attributes of the dataset

        Returns
        ----------
         numbers of attributes: int
            Numbers of attributes of the dataset
        """
        return len(self.attributes)

    def set_labels(self, labels):
        """
        Set function for dataset labels name

        Parameters
        ----------
        labels: List of string
            List of the labels name of the dataset
        """
        self.labels = labels

    def get_labels(self):
        """
        Get function for dataset labels name

        Returns
        ----------
        labels : List of strings
            Labels name of the dataset
        """
        return self.labels

    def get_number_labels(self):
        """
        Get numbers of labels of the dataset

        Returns
        ----------
        numbers of labels: int
            Numbers of labels of the dataset
        """
        count = 0
        for x in self.attributes.keys():
            if self.attributes[x] == 1:
                count += 1
        return count

    def get_bag(self, key):
        """
        Get data of a bag of the dataset

        Returns
        ----------
        bag: ndarray
            Attributes and labels of a bag of the dataset
        """
        # TODO: Formatearlo para que se vea bonito
        # TODO: Hacerlo quizas en funcion print_bag
        return self.data[key]

    def add_bag(self, bag: Bag):
        """
        Add a bag to the dataset

        Parameters
        ----------
        bag : Bag
            Bag to be added

        """
        self.data[bag.key] = bag

    def show_bag(self, key):
        # TODO: Check
        bag = self.get_bag(key)
        table = [[key] + self.get_attributes() + (self.get_labels())]
        count = 0
        for instance in bag[0]:
            table.append([count] + list(instance) + (list(bag[1])))
            count += 1
        # table = [['col 1', 'col 2', 'col 3', 'col 4'], [1, 2222, 30, 500], [4, 55, 6777, 1]]
        # print(tabulate(table, headers='firstrow', tablefmt='fancy_grid'))
        # print(tabulate([key], tablefmt="grid"))
        print(tabulate(table, headers='firstrow', tablefmt="grid", numalign="center"))

    def get_instance(self, key, index):
        """

        Parameters
        ----------
        key : string
            Key of the bag
            
        index : int
            Index of the instance in the bag

        Returns
        -------
        instance : tuple of ndarrays
            Tuple with attribute values and label of the instance

        """
        # TODO: check
        return [self.get_bag(key)[0][index], self.get_bag(key)[1]]

    def add_instance(self, key, values):
        """

        Parameters
        ----------
        key : string
            Key of the bag
        values : ndarray
            Values of the instance to be inserted

        """

        instance = np.array(values, ndmin=2)
        self.data[key] = [np.append(self.get_bag(key)[0], instance, axis=0), self.data[key][1]]
        # TODO: Optimizarlo

    def get_number_bags(self):
        """
        Get numbers of bags of the dataset

        Returns
        ----------
        numbers of bags: int
            Numbers of bags of the dataset
        """
        return len(self.data)

    def get_number_instances(self, key):

        # for x in all bags: sum of x.get_number_instances
        pass

    def set_attribute(self, key, attribute, value):
        """
            Update value from attributes

            Parameters
            ----------
            key : string
                Bag key of the dataset

            attribute: string
                Attribute of the dataset

            value: float
                New value for the update
            """
        # TODO: Implementarlo
        pass

    def show_dataset(self, head=None, attributes=None, labels=None):
        """"
        Function to show information about the dataset
        """
        # TODO: Formatearlo para que se vea bonito
        # TODO: Hacer algo como head y tail de pandas, ponerlo como parametro quizas, tambien lista atributos y labels
        #  a mostrar opcionales
        print("Name: ", self.get_name())
        print("Attributes: ", self.get_attributes())
        print("Labels: ", self.get_labels())
        print("Bags:")
        count = 0
        for key in self.data:
            # print("\n")
            bag = self.get_bag(key)
            # print("Key: ", key)
            # print("Attributes: ", bag[0])
            # print("Labels: ", bag[1])
            bag.show_bag()
            count += 1
            if head is not None:
                if count >= head:
                    break
            # print(bag)

    def cardinality(self):
        """
        Computes the Cardinality as the average number of labels per pattern.

        Returns
        ----------
        cardinality : float
            Average number of labels per pattern
        """
        suma = 0
        for key in self.data:
            suma += sum(self.data[key][1])
        return suma / len(self.data)

    def density(self):
        """
        Computes the density as the cardinality / numLabels.

        Returns
        ----------
        density : float
            Cardinality divived by number of labels
        """
        return self.cardinality() / self.get_number_labels()

    def distinct(self):
        """
        Computes the numbers of labels combinations used in the dataset respect all the possible ones
        
        Returns
        -------
        distinct : float
            Numbers of labels combinations used in the dataset divided by all possible combinations
        """
        options = set()
        for key in self.data:
            options.add(tuple(self.data[key][1]))
        return len(options) / (2 ** self.get_number_labels())

    def get_statistics(self):
        n_instances = 0
        max_instances = 0
        # TODO: check
        min_instances = float("inf")
        distribution = dict()
        for key in self.data:
            instances_bag = len(self.get_bag(key)[0])
            n_instances += instances_bag
            if instances_bag in distribution:
                distribution[instances_bag] += 1
            else:
                distribution[instances_bag] = 1
            if instances_bag < min_instances:
                min_instances = instances_bag
            elif instances_bag > max_instances:
                max_instances = instances_bag
        return n_instances, min_instances, max_instances, distribution

    def describe(self):
        """
        Print statistics about the dataset
        """

        # TODO: Ponerlo bonito con tabulate

        print("-----MULTILABEL-----")
        print("Cardinalidad: ", self.cardinality())
        print("Densidad: ", self.density())
        print("Distinct: ", self.distinct())
        print("")
        # TODO: Testearlo
        n_instances, min_instances, max_instances, distribution = self.get_statistics()
        print("-----MULTIINSTANCE-----")
        print("Nº of bags: ", self.get_number_bags())
        print("Total instances: ", n_instances)
        print("Average Instances per bag: ", n_instances / self.get_number_bags())
        print("Min Instances per bag: ", min_instances)
        print("Max Instances per bag: ", max_instances)
        print("Attributes per bag: ", self.get_number_attributes())
        # TODO: Implementarlo
        # sb.append("\nDistribution of bags <nBags, nInstances>:");
